﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Lambdas
{
    class Program
    {
        // functions can take other functions/actions as parameters
        static void F(Action a, Func<double, double> f)
        {
            a();
            Console.WriteLine("Called f(25) from F with the result {0}", f(25));
        }

        static void Main(string[] args)
        {
            // there are 2 basic types of lambda expressions:
            // Action and Func
            // Action repsent a method
            // Func represents a function with a return value

            // Action without an argument
            Action a1 = () => Console.WriteLine("a1 called");
            // Action with a single argument of type int
            Action<int> a2 = x => Console.WriteLine("a2 called on {0}", x);
            // An action taking 2 string arguments with with it's own code block
            Action<string, string> a3 =
                (l, r) =>
                {
                    if (l.Equals(r)) Console.WriteLine("strings are equal");
                    else Console.WriteLine("strings are not equal");
                };

            // Type of the parameters can also be specified in the "argument list". This is useful for certain LINQ expressions (can't think of any :)) or more clarity
            Action<char, int> a4 = (char s, int n) => Console.WriteLine(new string(s, n));

            // A function returning a value/object
            Func<int> f1 = () => 42;
            // A function taking a single argument and returning it's square
            Func<double, double> f2 = x => x * x;
            // A function that concatenates 2 strings
            Func<string, string, string> f3 = (l, r) => l + r;
            // A function that returns an Action
            // in other words - we put lambdas in your lambdas so that you can lambda while you lambda :)
            Func<int, Action<char>> f4 = x => (y => Console.WriteLine(new string(y, x)));
            // If a lambda has a "body", it has to return a value
            Func<double, double> f5 = x =>
                {
                    double y = x - 10;
                    return y;
                };

            // lambdas are called as normal functions
            a1();
            a2(10);
            a3("a", "b");
            a4('c', 10);

            int f1result = f1();
            double f2result = f2(Math.PI);
            string f3result = f3("left-- ", " --right");
            f4(10)('f'); // cool, eh...
            double f5result = f5(1.123);

            // you can also call a lambda like this
            a2.Invoke(123);
            // or asynchronously using Being/End invoke - more on this later

            F(a1, f2);
            F(() => Console.WriteLine("lambda passed as a function parameter"), x => x + 1);

            Console.ReadLine();
        }
    }
}
